@@ -2,10 +2,24 @@
 import argparse
 import base64
 import sys
+from typing import (
+    Any,
+    Dict,
+    FrozenSet,
+    List,
+    Mapping,
+    Optional,
+    Tuple,
+    Type,
+    cast,
+)
 
-import OpenSSL
+from OpenSSL import crypto
 
-from josepy import b64, errors, json_util, jwa, jwk, util
+import josepy
+from josepy import b64, errors, json_util, jwa
+from josepy import jwk as jwk_mod
+from josepy import util
 
 
 class MediaType:
@@ -15,7 +29,7 @@ class MediaType:
     """MIME Media Type and Content Type prefix."""
 
     @classmethod
-    def decode(cls, value):
+    def decode(cls, value: str) -> str:
         """Decoder."""
         # 4.1.10
         if '/' not in value:
@@ -25,7 +39,7 @@ class MediaType:
         return value
 
     @classmethod
-    def encode(cls, value):
+    def encode(cls, value: str) -> str:
         """Encoder."""
         # 4.1.10
         if ';' not in value:
@@ -55,30 +69,32 @@ class Header(json_util.JSONObjectWithFields):
     :ivar str cty: Content-Type, inc. :const:`MediaType.PREFIX`.
 
     """
-    alg = json_util.Field(
+    alg: Optional[jwa.JWASignature] = json_util.field(
         'alg', decoder=jwa.JWASignature.from_json, omitempty=True)
-    jku = json_util.Field('jku', omitempty=True)
-    jwk = json_util.Field('jwk', decoder=jwk.JWK.from_json, omitempty=True)
-    kid = json_util.Field('kid', omitempty=True)
-    x5u = json_util.Field('x5u', omitempty=True)
-    x5c = json_util.Field('x5c', omitempty=True, default=())
-    x5t = json_util.Field(
+    jku: Optional[bytes] = json_util.field('jku', omitempty=True)
+    jwk: Optional[jwk_mod.JWK] = json_util.field(
+        'jwk', decoder=jwk_mod.JWK.from_json, omitempty=True)
+    kid: Optional[str] = json_util.field('kid', omitempty=True)
+    x5u: Optional[bytes] = json_util.field('x5u', omitempty=True)
+    x5c: Tuple[util.ComparableX509, ...] = json_util.field('x5c', omitempty=True, default=())
+    x5t: Optional[bytes] = json_util.field(
         'x5t', decoder=json_util.decode_b64jose, omitempty=True)
-    x5tS256 = json_util.Field(
+    x5tS256: Optional[bytes] = json_util.field(
         'x5t#S256', decoder=json_util.decode_b64jose, omitempty=True)
-    typ = json_util.Field('typ', encoder=MediaType.encode,
-                          decoder=MediaType.decode, omitempty=True)
-    cty = json_util.Field('cty', encoder=MediaType.encode,
-                          decoder=MediaType.decode, omitempty=True)
-    crit = json_util.Field('crit', omitempty=True, default=())
+    typ: Optional[MediaType] = json_util.field('typ', encoder=MediaType.encode,
+                                               decoder=MediaType.decode, omitempty=True)
+    cty: Optional[MediaType] = json_util.field('cty', encoder=MediaType.encode,
+                                               decoder=MediaType.decode, omitempty=True)
+    crit: Tuple[Any, ...] = json_util.field('crit', omitempty=True, default=())
+    _fields: Dict[str, json_util.Field]
 
-    def not_omitted(self):
+    def not_omitted(self) -> Dict[str, json_util.Field]:
         """Fields that would not be omitted in the JSON object."""
         return {name: getattr(self, name)
                 for name, field in self._fields.items()
                 if not field.omit(getattr(self, name))}
 
-    def __add__(self, other):
+    def __add__(self, other: Any) -> 'Header':
         if not isinstance(other, type(self)):
             raise TypeError('Header cannot be added to: {0}'.format(
                 type(other)))
@@ -90,9 +106,9 @@ class Header(json_util.JSONObjectWithFields):
             raise TypeError('Addition of overlapping headers not defined')
 
         not_omitted_self.update(not_omitted_other)
-        return type(self)(**not_omitted_self)  # pylint: disable=star-args
+        return type(self)(**not_omitted_self)
 
-    def find_key(self):
+    def find_key(self) -> josepy.JWK:
         """Find key based on header.
 
         .. todo:: Supports only "jwk" header parameter lookup.
@@ -108,25 +124,24 @@ class Header(json_util.JSONObjectWithFields):
         return self.jwk
 
     @crit.decoder  # type: ignore
-    def crit(unused_value):
-        # pylint: disable=missing-docstring,no-self-argument,no-self-use
+    def crit(unused_value: Any) -> Any:
         raise errors.DeserializationError(
             '"crit" is not supported, please subclass')
 
     # x5c does NOT use JOSE Base64 (4.1.6)
 
     @x5c.encoder  # type: ignore
-    def x5c(value):  # pylint: disable=missing-docstring,no-self-argument
-        return [base64.b64encode(OpenSSL.crypto.dump_certificate(
-            OpenSSL.crypto.FILETYPE_ASN1, cert.wrapped)) for cert in value]
+    def x5c(value):
+        return [base64.b64encode(crypto.dump_certificate(
+            crypto.FILETYPE_ASN1, cert.wrapped)) for cert in value]
 
     @x5c.decoder  # type: ignore
-    def x5c(value):  # pylint: disable=missing-docstring,no-self-argument
+    def x5c(value):
         try:
-            return tuple(util.ComparableX509(OpenSSL.crypto.load_certificate(
-                OpenSSL.crypto.FILETYPE_ASN1,
+            return tuple(util.ComparableX509(crypto.load_certificate(
+                crypto.FILETYPE_ASN1,
                 base64.b64decode(cert))) for cert in value)
-        except OpenSSL.crypto.Error as error:
+        except crypto.Error as error:
             raise errors.DeserializationError(error)
 
 
@@ -141,33 +156,34 @@ class Signature(json_util.JSONObjectWithFields):
 
     """
     header_cls = Header
+    combined: Header
 
     __slots__ = ('combined',)
-    protected = json_util.Field('protected', omitempty=True, default='')
-    header = json_util.Field(
+    protected: str = json_util.field('protected', omitempty=True, default='')
+    header: Header = json_util.field(
         'header', omitempty=True, default=header_cls(),
         decoder=header_cls.from_json)
-    signature = json_util.Field(
+    signature: bytes = json_util.field(
         'signature', decoder=json_util.decode_b64jose,
         encoder=json_util.encode_b64jose)
 
     @protected.encoder  # type: ignore
-    def protected(value):  # pylint: disable=missing-docstring,no-self-argument
+    def protected(value: str) -> str:
         # wrong type guess (Signature, not bytes) | pylint: disable=no-member
         return json_util.encode_b64jose(value.encode('utf-8'))
 
     @protected.decoder  # type: ignore
-    def protected(value):  # pylint: disable=missing-docstring,no-self-argument
+    def protected(value: str) -> str:
         return json_util.decode_b64jose(value).decode('utf-8')
 
-    def __init__(self, **kwargs):
+    def __init__(self, **kwargs: Any) -> None:
         if 'combined' not in kwargs:
             kwargs = self._with_combined(kwargs)
         super().__init__(**kwargs)
         assert self.combined.alg is not None
 
     @classmethod
-    def _with_combined(cls, kwargs):
+    def _with_combined(cls, kwargs: Any) -> Dict[str, Any]:
         assert 'combined' not in kwargs
         header = kwargs.get('header', cls._fields['header'].default)
         protected = kwargs.get('protected', cls._fields['protected'].default)
@@ -181,27 +197,35 @@ class Signature(json_util.JSONObjectWithFields):
         return kwargs
 
     @classmethod
-    def _msg(cls, protected, payload):
+    def _msg(cls, protected: str, payload: bytes) -> bytes:
         return (b64.b64encode(protected.encode('utf-8')) + b'.' +
                 b64.b64encode(payload))
 
-    def verify(self, payload, key=None):
+    def verify(self, payload: bytes, key: Optional[josepy.JWK] = None) -> bool:
         """Verify.
 
+        :param bytes payload: Payload to verify.
         :param JWK key: Key used for verification.
 
         """
-        key = self.combined.find_key() if key is None else key
+        actual_key: josepy.JWK = self.combined.find_key() if key is None else key
+        if not self.combined.alg:
+            raise josepy.Error("Not signature algorithm defined.")
         return self.combined.alg.verify(
-            key=key.key, sig=self.signature,
+            key=actual_key.key, sig=self.signature,
             msg=self._msg(self.protected, payload))
 
     @classmethod
-    def sign(cls, payload, key, alg, include_jwk=True,
-             protect=frozenset(), **kwargs):
+    def sign(cls, payload: bytes, key: josepy.JWK, alg: josepy.JWASignature,
+             include_jwk: bool = True, protect: FrozenSet = frozenset(),
+             **kwargs: Any) -> 'Signature':
         """Sign.
 
+        :param bytes payload: Payload to sign.
         :param JWK key: Key for signature.
+        :param JWASignature alg: Signature algorithm to use to sign.
+        :param bool include_jwk: If True, insert the JWK inside the signature headers.
+        :param FrozenSet protect: List of headers to protect.
 
         """
         assert isinstance(key, alg.kty)
@@ -219,24 +243,23 @@ class Signature(json_util.JSONObjectWithFields):
             if header in header_params:
                 protected_params[header] = header_params.pop(header)
         if protected_params:
-            # pylint: disable=star-args
             protected = cls.header_cls(**protected_params).json_dumps()
         else:
             protected = ''
 
-        header = cls.header_cls(**header_params)  # pylint: disable=star-args
+        header = cls.header_cls(**header_params)
         signature = alg.sign(key.key, cls._msg(protected, payload))
 
         return cls(protected=protected, header=header, signature=signature)
 
-    def fields_to_partial_json(self):
+    def fields_to_partial_json(self) -> Dict[str, Any]:
         fields = super().fields_to_partial_json()
         if not fields['header'].not_omitted():
             del fields['header']
         return fields
 
     @classmethod
-    def fields_from_json(cls, jobj):
+    def fields_from_json(cls, jobj: Mapping[str, Any]) -> Dict[str, Any]:
         fields = super().fields_from_json(jobj)
         fields_with_combined = cls._with_combined(fields)
         if 'alg' not in fields_with_combined['combined'].not_omitted():
@@ -252,21 +275,23 @@ class JWS(json_util.JSONObjectWithFields):
 
     """
     __slots__ = ('payload', 'signatures')
+    payload: bytes
+    signatures: List[Signature]
 
     signature_cls = Signature
 
-    def verify(self, key=None):
+    def verify(self, key: Optional[josepy.JWK] = None) -> bool:
         """Verify."""
         return all(sig.verify(self.payload, key) for sig in self.signatures)
 
     @classmethod
-    def sign(cls, payload, **kwargs):
+    def sign(cls, payload: bytes, **kwargs: Any) -> 'JWS':
         """Sign."""
         return cls(payload=payload, signatures=(
             cls.signature_cls.sign(payload=payload, **kwargs),))
 
     @property
-    def signature(self):
+    def signature(self) -> Signature:
         """Get a singleton signature.
 
         :rtype: :class:`JWS.signature_cls`
@@ -275,7 +300,7 @@ class JWS(json_util.JSONObjectWithFields):
         assert len(self.signatures) == 1
         return self.signatures[0]
 
-    def to_compact(self):
+    def to_compact(self) -> bytes:
         """Compact serialization.
 
         :rtype: bytes
@@ -294,7 +319,7 @@ class JWS(json_util.JSONObjectWithFields):
             b64.b64encode(self.signature.signature))
 
     @classmethod
-    def from_compact(cls, compact):
+    def from_compact(cls, compact: bytes) -> 'JWS':
         """Compact deserialization.
 
         :param bytes compact:
@@ -312,7 +337,7 @@ class JWS(json_util.JSONObjectWithFields):
             signature=b64.b64decode(signature))
         return cls(payload=b64.b64decode(payload), signatures=(sig,))
 
-    def to_partial_json(self, flat=True):  # pylint: disable=arguments-differ
+    def to_partial_json(self, flat: bool = True) -> Dict[str, Any]:
         assert self.signatures
         payload = json_util.encode_b64jose(self.payload)
 
@@ -327,12 +352,13 @@ class JWS(json_util.JSONObjectWithFields):
             }
 
     @classmethod
-    def from_json(cls, jobj):
+    def from_json(cls, jobj: Mapping[str, Any]) -> 'JWS':
         if 'signature' in jobj and 'signatures' in jobj:
             raise errors.DeserializationError('Flat mixed with non-flat')
         elif 'signature' in jobj:  # flat
-            return cls(payload=json_util.decode_b64jose(jobj.pop('payload')),
-                       signatures=(cls.signature_cls.from_json(jobj),))
+            filtered = {key: value for key, value in jobj.items() if key != "payload"}
+            return cls(payload=json_util.decode_b64jose(jobj['payload']),
+                       signatures=(cls.signature_cls.from_json(filtered),))
         else:
             return cls(payload=json_util.decode_b64jose(jobj['payload']),
                        signatures=tuple(cls.signature_cls.from_json(sig)
@@ -343,7 +369,7 @@ class CLI:
     """JWS CLI."""
 
     @classmethod
-    def sign(cls, args):
+    def sign(cls, args: argparse.Namespace) -> None:
         """Sign."""
         key = args.alg.kty.load(args.key.read())
         args.key.close()
@@ -361,16 +387,16 @@ class CLI:
             print(sig.json_dumps_pretty())
 
     @classmethod
-    def verify(cls, args):
+    def verify(cls, args: argparse.Namespace) -> bool:
         """Verify."""
         if args.compact:
             sig = JWS.from_compact(sys.stdin.read().encode())
         else:  # JSON
             try:
-                sig = JWS.json_loads(sys.stdin.read())
+                sig = cast(JWS, JWS.json_loads(sys.stdin.read()))
             except errors.Error as error:
                 print(error)
-                return -1
+                return False
 
         if args.key is not None:
             assert args.kty is not None
@@ -379,25 +405,25 @@ class CLI:
         else:
             key = None
 
-        sys.stdout.write(sig.payload)
+        sys.stdout.write(sig.payload.decode())
         return not sig.verify(key=key)
 
     @classmethod
-    def _alg_type(cls, arg):
+    def _alg_type(cls, arg: Any) -> jwa.JWASignature:
         return jwa.JWASignature.from_json(arg)
 
     @classmethod
-    def _header_type(cls, arg):
+    def _header_type(cls, arg: Any) -> Any:
         assert arg in Signature.header_cls._fields
         return arg
 
     @classmethod
-    def _kty_type(cls, arg):
-        assert arg in jwk.JWK.TYPES
-        return jwk.JWK.TYPES[arg]
+    def _kty_type(cls, arg: Any) -> Type[jwk_mod.JWK]:
+        assert arg in jwk_mod.JWK.TYPES
+        return jwk_mod.JWK.TYPES[arg]
 
     @classmethod
-    def run(cls, args=None):
+    def run(cls, args: List[str] = None) -> Optional[bool]:
         """Parse arguments and sign/verify."""
         if args is None:
             args = sys.argv[1:]
