@@ -9,15 +9,33 @@ The framework presented here is somewhat based on `Go's "json" package`_
 import abc
 import binascii
 import logging
-from typing import Dict, Type
+from typing import Any, Callable, Dict, List, Mapping, Optional, Type, TypeVar
 
-import OpenSSL
+from OpenSSL import crypto
 
 from josepy import b64, errors, interfaces, util
 
 logger = logging.getLogger(__name__)
 
 
+def field(json_name: str, default: Any = None, omitempty: bool = False,
+          decoder: Callable[[Any], Any] = None, encoder: Callable[[Any], Any] = None) -> Any:
+    """Convenient function to declare a :class:`Field` with proper type annotations.
+
+    This function allows to write the following code:
+
+    import josepy
+    class JSON(josepy.JSONObjectWithFields):
+        typ: str = josepy.field('type')
+
+        def other_type(self) -> str:
+            return self.typ
+
+    """
+    return _TypedField(json_name=json_name, default=default, omitempty=omitempty,
+                       decoder=decoder, encoder=encoder)
+
+
 class Field:
     """JSON object field.
 
@@ -44,9 +62,9 @@ class Field:
     """
     __slots__ = ('json_name', 'default', 'omitempty', 'fdec', 'fenc')
 
-    def __init__(self, json_name, default=None, omitempty=False,
-                 decoder=None, encoder=None):
-        # pylint: disable=too-many-arguments
+    def __init__(self, json_name: str, default: Any = None, omitempty: bool = False,
+                 decoder: Callable[[Any], Any] = None,
+                 encoder: Callable[[Any], Any] = None) -> None:
         self.json_name = json_name
         self.default = default
         self.omitempty = omitempty
@@ -55,7 +73,7 @@ class Field:
         self.fenc = self.default_encoder if encoder is None else encoder
 
     @classmethod
-    def _empty(cls, value):
+    def _empty(cls, value: Any) -> bool:
         """Is the provided value considered "empty" for this field?
 
         This is useful for subclasses that might want to override the
@@ -64,37 +82,39 @@ class Field:
         """
         return not isinstance(value, bool) and not value
 
-    def omit(self, value):
+    def omit(self, value: Any) -> bool:
         """Omit the value in output?"""
         return self._empty(value) and self.omitempty
 
-    def _update_params(self, **kwargs):
+    def _update_params(self, **kwargs: Any) -> 'Field':
         current = {
-            "json_name": self.json_name, "default": self.default,
+            "json_name": self.json_name,
+            "default": self.default,
             "omitempty": self.omitempty,
-            "decoder": self.fdec, "encoder": self.fenc,
+            "decoder": self.fdec,
+            "encoder": self.fenc,
             **kwargs,
         }
-        return type(self)(**current)  # pylint: disable=star-args
+        return type(self)(**current)  # type: ignore[arg-type]
 
-    def decoder(self, fdec):
+    def decoder(self, fdec: Callable[[Any], Any]) -> 'Field':
         """Descriptor to change the decoder on JSON object field."""
         return self._update_params(decoder=fdec)
 
-    def encoder(self, fenc):
+    def encoder(self, fenc: Callable[[Any], Any]) -> 'Field':
         """Descriptor to change the encoder on JSON object field."""
         return self._update_params(encoder=fenc)
 
-    def decode(self, value):
+    def decode(self, value: Any) -> Any:
         """Decode a value, optionally with context JSON object."""
         return self.fdec(value)
 
-    def encode(self, value):
+    def encode(self, value: Any) -> Any:
         """Encode a value, optionally with context JSON object."""
         return self.fenc(value)
 
     @classmethod
-    def default_decoder(cls, value):
+    def default_decoder(cls, value: Any) -> Any:
         """Default decoder.
 
         Recursively deserialize into immutable types (
@@ -113,13 +133,22 @@ class Field:
             return value
 
     @classmethod
-    def default_encoder(cls, value):
+    def default_encoder(cls, value: Any) -> Any:
         """Default (passthrough) encoder."""
         # field.to_partial_json() is no good as encoder has to do partial
         # serialization only
         return value
 
 
+class _TypedField(Field):
+    """Specialized class to mark a JSON object field with typed annotations.
+
+    This class is kept private because fields are supposed to be declared
+    using the :function:`field` in this situation.
+
+    In the future the :class:`Field` may be removed in favor of this one."""
+
+
 class JSONObjectWithFieldsMeta(abc.ABCMeta):
     """Metaclass for :class:`JSONObjectWithFields` and its subclasses.
 
@@ -162,30 +191,38 @@ class JSONObjectWithFieldsMeta(abc.ABCMeta):
 
     _fields: Dict[str, Field] = {}
 
-    def __new__(mcs, name, bases, dikt):
+    def __new__(mcs, name: str, bases: List[str],
+                namespace: Dict[str, Any]) -> 'JSONObjectWithFieldsMeta':
         fields = {}
 
         for base in bases:
             fields.update(getattr(base, '_fields', {}))
         # Do not reorder, this class might override fields from base classes!
-        # We create a tuple based on dikt.items() here because the loop
-        # modifies dikt.
-        for key, value in tuple(dikt.items()):
+        # We use a copy of namespace in the loop because the loop modifies it.
+        for key, value in namespace.copy().items():
             if isinstance(value, Field):
-                fields[key] = dikt.pop(key)
+                if isinstance(value, _TypedField):
+                    # Ensure the type annotation has been set for the field.
+                    # Error out if it is not the case.
+                    if key not in namespace.get('__annotations__', {}):
+                        raise ValueError(
+                            f'Field `{key}` in JSONObject `{name}` has no type annotation.')
+                fields[key] = namespace.pop(key)
 
-        dikt['_orig_slots'] = dikt.get('__slots__', ())
-        dikt['__slots__'] = tuple(
-            list(dikt['_orig_slots']) + list(fields.keys()))
-        dikt['_fields'] = fields
+        namespace['_orig_slots'] = namespace.get('__slots__', ())
+        namespace['__slots__'] = tuple(
+            list(namespace['_orig_slots']) + list(fields.keys()))
+        namespace['_fields'] = fields
 
-        return abc.ABCMeta.__new__(mcs, name, bases, dikt)
+        return abc.ABCMeta.__new__(mcs, name, bases, namespace)  # type: ignore[call-overload]
+
+
+GenericJSONObjectWithFields = TypeVar('GenericJSONObjectWithFields', bound='JSONObjectWithFields')
 
 
 class JSONObjectWithFields(util.ImmutableMap,
                            interfaces.JSONDeSerializable,
                            metaclass=JSONObjectWithFieldsMeta):
-    # pylint: disable=too-few-public-methods
     """JSON object with fields.
 
     Example::
@@ -211,20 +248,17 @@ class JSONObjectWithFields(util.ImmutableMap,
       assert Foo(bar='baz').bar == 'baz'
 
     """
-
     @classmethod
-    def _defaults(cls) -> Dict:
+    def _defaults(cls) -> Dict[str, Any]:
         """Get default fields values."""
         return {
             slot: field.default for slot, field in cls._fields.items()
         }
 
-    def __init__(self, **kwargs):
-        # pylint: disable=star-args
-        super().__init__(
-            **{**self._defaults(), **kwargs})
+    def __init__(self, **kwargs: Any) -> None:
+        super().__init__(**{**self._defaults(), **kwargs})
 
-    def encode(self, name):
+    def encode(self, name: str) -> Any:
         """Encode a single field.
 
         :param str name: Name of the field to be encoded.
@@ -234,17 +268,17 @@ class JSONObjectWithFields(util.ImmutableMap,
 
         """
         try:
-            field = self._fields[name]
+            field = self._fields[name]  # type: ignore[attr-defined]
         except KeyError:
             raise errors.Error("Field not found: {0}".format(name))
 
         return field.encode(getattr(self, name))
 
-    def fields_to_partial_json(self):
+    def fields_to_partial_json(self) -> Dict[str, Any]:
         """Serialize fields to JSON."""
         jobj = {}
         omitted = set()
-        for slot, field in self._fields.items():
+        for slot, field in self._fields.items():  # type: ignore[attr-defined]
             value = getattr(self, slot)
 
             if field.omit(value):
@@ -258,11 +292,11 @@ class JSONObjectWithFields(util.ImmutableMap,
                             slot, value, error))
         return jobj
 
-    def to_partial_json(self):
+    def to_partial_json(self) -> Dict[str, Any]:
         return self.fields_to_partial_json()
 
     @classmethod
-    def _check_required(cls, jobj):
+    def _check_required(cls, jobj: Mapping[str, Any]) -> None:
         missing = set()
         for _, field in cls._fields.items():
             if not field.omitempty and field.json_name not in jobj:
@@ -274,7 +308,7 @@ class JSONObjectWithFields(util.ImmutableMap,
                     ','.join(missing)))
 
     @classmethod
-    def fields_from_json(cls, jobj):
+    def fields_from_json(cls, jobj: Mapping[str, Any]) -> Any:
         """Deserialize fields from JSON."""
         cls._check_required(jobj)
         fields = {}
@@ -286,13 +320,14 @@ class JSONObjectWithFields(util.ImmutableMap,
                 try:
                     fields[slot] = field.decode(value)
                 except errors.DeserializationError as error:
-                    raise errors.DeserializationError(
-                        'Could not decode {0!r} ({1!r}): {2}'.format(
-                            slot, value, error))
+                    raise errors.DeserializationError('Could not decode {0!r} ({1!r}): {2}'.format(
+                        slot, value, error))
         return fields
 
     @classmethod
-    def from_json(cls, jobj):
+    def from_json(
+        cls: Type[GenericJSONObjectWithFields], jobj: Mapping[str, Any]
+    ) -> GenericJSONObjectWithFields:
         return cls(**cls.fields_from_json(jobj))
 
 
@@ -307,7 +342,7 @@ def encode_b64jose(data: bytes) -> str:
     return b64.b64encode(data).decode('ascii')
 
 
-def decode_b64jose(data, size=None, minimum=False):
+def decode_b64jose(data: str, size: Optional[int] = None, minimum: bool = False) -> bytes:
     """Decode JOSE Base-64 field.
 
     :param unicode data:
@@ -331,7 +366,7 @@ def decode_b64jose(data, size=None, minimum=False):
     return decoded
 
 
-def encode_hex16(value):
+def encode_hex16(value: bytes) -> str:
     """Hexlify.
 
     :param bytes value:
@@ -341,7 +376,7 @@ def encode_hex16(value):
     return binascii.hexlify(value).decode()
 
 
-def decode_hex16(value, size=None, minimum=False):
+def decode_hex16(value: str, size: Optional[int] = None, minimum: bool = False) -> bytes:
     """Decode hexlified field.
 
     :param unicode value:
@@ -352,28 +387,31 @@ def decode_hex16(value, size=None, minimum=False):
     :rtype: bytes
 
     """
-    value = value.encode()
-    if size is not None and ((not minimum and len(value) != size * 2) or
-                             (minimum and len(value) < size * 2)):
+    value_b = value.encode()
+    if size is not None and ((not minimum and len(value_b) != size * 2) or
+                             (minimum and len(value_b) < size * 2)):
         raise errors.DeserializationError()
     try:
-        return binascii.unhexlify(value)
+        return binascii.unhexlify(value_b)
     except binascii.Error as error:
         raise errors.DeserializationError(error)
 
 
-def encode_cert(cert):
+def encode_cert(cert: util.ComparableX509) -> str:
     """Encode certificate as JOSE Base-64 DER.
 
     :type cert: `OpenSSL.crypto.X509` wrapped in `.ComparableX509`
     :rtype: unicode
 
     """
-    return encode_b64jose(OpenSSL.crypto.dump_certificate(
-        OpenSSL.crypto.FILETYPE_ASN1, cert.wrapped))
+    if isinstance(cert.wrapped, crypto.X509Req):
+        raise ValueError("Error input is actually a certificate request.")
+
+    return encode_b64jose(crypto.dump_certificate(
+        crypto.FILETYPE_ASN1, cert.wrapped))
 
 
-def decode_cert(b64der):
+def decode_cert(b64der: str) -> util.ComparableX509:
     """Decode JOSE Base-64 DER-encoded certificate.
 
     :param unicode b64der:
@@ -381,24 +419,27 @@ def decode_cert(b64der):
 
     """
     try:
-        return util.ComparableX509(OpenSSL.crypto.load_certificate(
-            OpenSSL.crypto.FILETYPE_ASN1, decode_b64jose(b64der)))
-    except OpenSSL.crypto.Error as error:
+        return util.ComparableX509(crypto.load_certificate(
+            crypto.FILETYPE_ASN1, decode_b64jose(b64der)))
+    except crypto.Error as error:
         raise errors.DeserializationError(error)
 
 
-def encode_csr(csr):
+def encode_csr(csr: util.ComparableX509) -> str:
     """Encode CSR as JOSE Base-64 DER.
 
     :type csr: `OpenSSL.crypto.X509Req` wrapped in `.ComparableX509`
     :rtype: unicode
 
     """
-    return encode_b64jose(OpenSSL.crypto.dump_certificate_request(
-        OpenSSL.crypto.FILETYPE_ASN1, csr.wrapped))
+    if isinstance(csr.wrapped, crypto.X509):
+        raise ValueError("Error input is actually a certificate.")
+
+    return encode_b64jose(crypto.dump_certificate_request(
+        crypto.FILETYPE_ASN1, csr.wrapped))
 
 
-def decode_csr(b64der):
+def decode_csr(b64der: str) -> util.ComparableX509:
     """Decode JOSE Base-64 DER-encoded CSR.
 
     :param unicode b64der:
@@ -406,12 +447,16 @@ def decode_csr(b64der):
 
     """
     try:
-        return util.ComparableX509(OpenSSL.crypto.load_certificate_request(
-            OpenSSL.crypto.FILETYPE_ASN1, decode_b64jose(b64der)))
-    except OpenSSL.crypto.Error as error:
+        return util.ComparableX509(crypto.load_certificate_request(
+            crypto.FILETYPE_ASN1, decode_b64jose(b64der)))
+    except crypto.Error as error:
         raise errors.DeserializationError(error)
 
 
+GenericTypedJSONObjectWithFields = TypeVar(
+    'GenericTypedJSONObjectWithFields', bound='TypedJSONObjectWithFields')
+
+
 class TypedJSONObjectWithFields(JSONObjectWithFields):
     """JSON object with type."""
 
@@ -429,14 +474,15 @@ class TypedJSONObjectWithFields(JSONObjectWithFields):
     """Types registered for JSON deserialization"""
 
     @classmethod
-    def register(cls, type_cls, typ=None):
+    def register(cls, type_cls: Type[GenericTypedJSONObjectWithFields],
+                 typ: Optional[str] = None) -> Type[GenericTypedJSONObjectWithFields]:
         """Register class for JSON deserialization."""
         typ = type_cls.typ if typ is None else typ
         cls.TYPES[typ] = type_cls
         return type_cls
 
     @classmethod
-    def get_type_cls(cls, jobj):
+    def get_type_cls(cls, jobj: Mapping[str, Any]) -> Type["TypedJSONObjectWithFields"]:
         """Get the registered class for ``jobj``."""
         if cls in cls.TYPES.values():
             if cls.type_field_name not in jobj:
@@ -460,7 +506,7 @@ class TypedJSONObjectWithFields(JSONObjectWithFields):
         except KeyError:
             raise errors.UnrecognizedTypeError(typ, jobj)
 
-    def to_partial_json(self):
+    def to_partial_json(self) -> Dict[str, Any]:
         """Get JSON serializable object.
 
         :returns: Serializable JSON object representing ACME typed object.
@@ -474,7 +520,7 @@ class TypedJSONObjectWithFields(JSONObjectWithFields):
         return jobj
 
     @classmethod
-    def from_json(cls, jobj):
+    def from_json(cls, jobj: Mapping[str, Any]) -> "TypedJSONObjectWithFields":
         """Deserialize ACME object from valid JSON object.
 
         :raises josepy.errors.UnrecognizedTypeError: if type
