@@ -5,25 +5,25 @@ https://tools.ietf.org/html/draft-ietf-jose-json-web-algorithms-40
 """
 import abc
 import logging
-from typing import Dict, Type
+from collections.abc import Hashable
+from typing import Any, Callable, Dict
 
 import cryptography.exceptions
 from cryptography.hazmat.backends import default_backend
-from cryptography.hazmat.primitives import hashes
-from cryptography.hazmat.primitives import hmac
-from cryptography.hazmat.primitives.asymmetric import padding, ec
-from cryptography.hazmat.primitives.asymmetric.utils import decode_dss_signature
-from cryptography.hazmat.primitives.asymmetric.utils import encode_dss_signature
+from cryptography.hazmat.primitives import hashes, hmac
+from cryptography.hazmat.primitives.asymmetric import ec, padding, rsa
+from cryptography.hazmat.primitives.asymmetric.utils import (
+    decode_dss_signature,
+    encode_dss_signature,
+)
+from cryptography.hazmat.primitives.hashes import HashAlgorithm
 
 from josepy import errors, interfaces, jwk
 
-from collections.abc import Hashable
-
 logger = logging.getLogger(__name__)
 
 
-class JWA(interfaces.JSONDeSerializable):  # pylint: disable=abstract-method
-    # pylint: disable=too-few-public-methods
+class JWA(interfaces.JSONDeSerializable):
     # for some reason disable=abstract-method has to be on the line
     # above...
     """JSON Web Algorithm."""
@@ -31,59 +31,60 @@ class JWA(interfaces.JSONDeSerializable):  # pylint: disable=abstract-method
 
 class JWASignature(JWA, Hashable):
     """Base class for JSON Web Signature Algorithms."""
-    SIGNATURES: Dict[str, Type] = {}
+    SIGNATURES: Dict[str, 'JWASignature'] = {}
+    kty: Any
 
-    def __init__(self, name):
+    def __init__(self, name: str) -> None:
         self.name = name
 
-    def __eq__(self, other):
+    def __eq__(self, other: Any) -> bool:
         if not isinstance(other, JWASignature):
             return NotImplemented
         return self.name == other.name
 
-    def __hash__(self):
+    def __hash__(self) -> int:
         return hash((self.__class__, self.name))
 
     @classmethod
-    def register(cls, signature_cls):
+    def register(cls, signature_cls: 'JWASignature') -> 'JWASignature':
         """Register class for JSON deserialization."""
         cls.SIGNATURES[signature_cls.name] = signature_cls
         return signature_cls
 
-    def to_partial_json(self):
+    def to_partial_json(self) -> Any:
         return self.name
 
     @classmethod
-    def from_json(cls, jobj):
+    def from_json(cls, jobj: Any) -> 'JWASignature':
         return cls.SIGNATURES[jobj]
 
     @abc.abstractmethod
-    def sign(self, key, msg):  # pragma: no cover
+    def sign(self, key: Any, msg: bytes) -> bytes:  # pragma: no cover
         """Sign the ``msg`` using ``key``."""
         raise NotImplementedError()
 
     @abc.abstractmethod
-    def verify(self, key, msg, sig):  # pragma: no cover
+    def verify(self, key: Any, msg: bytes, sig: bytes) -> bool:  # pragma: no cover
         """Verify the ``msg`` and ``sig`` using ``key``."""
         raise NotImplementedError()
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return self.name
 
 
 class _JWAHS(JWASignature):
     kty = jwk.JWKOct
 
-    def __init__(self, name, hash_):
+    def __init__(self, name: str, hash_: Callable[[], HashAlgorithm]):
         super().__init__(name)
         self.hash = hash_()
 
-    def sign(self, key, msg):
+    def sign(self, key: bytes, msg: bytes) -> bytes:
         signer = hmac.HMAC(key, self.hash, backend=default_backend())
         signer.update(msg)
         return signer.finalize()
 
-    def verify(self, key, msg, sig):
+    def verify(self, key: bytes, msg: bytes, sig: bytes) -> bool:
         verifier = hmac.HMAC(key, self.hash, backend=default_backend())
         verifier.update(msg)
         try:
@@ -97,42 +98,24 @@ class _JWAHS(JWASignature):
 
 class _JWARSA:
     kty = jwk.JWKRSA
-    padding = NotImplemented
-    hash = NotImplemented
+    padding: Any = NotImplemented
+    hash: HashAlgorithm = NotImplemented
 
-    def sign(self, key, msg):
+    def sign(self, key: rsa.RSAPrivateKey, msg: bytes) -> bytes:
         """Sign the ``msg`` using ``key``."""
-        # If cryptography library supports new style api (v1.4 and later)
-        new_api = hasattr(key, "sign")
         try:
-            if new_api:
-                return key.sign(msg, self.padding, self.hash)
-            signer = key.signer(self.padding, self.hash)
+            return key.sign(msg, self.padding, self.hash)
         except AttributeError as error:
             logger.debug(error, exc_info=True)
             raise errors.Error("Public key cannot be used for signing")
         except ValueError as error:  # digest too large
             logger.debug(error, exc_info=True)
             raise errors.Error(str(error))
-        signer.update(msg)
-        try:
-            return signer.finalize()
-        except ValueError as error:
-            logger.debug(error, exc_info=True)
-            raise errors.Error(str(error))
 
-    def verify(self, key, msg, sig):
+    def verify(self, key: rsa.RSAPublicKey, msg: bytes, sig: bytes) -> bool:
         """Verify the ``msg` and ``sig`` using ``key``."""
-        # If cryptography library supports new style api (v1.4 and later)
-        new_api = hasattr(key, "verify")
-        if not new_api:
-            verifier = key.verifier(sig, self.padding, self.hash)
-            verifier.update(msg)
         try:
-            if new_api:
-                key.verify(sig, msg, self.padding, self.hash)
-            else:
-                verifier.verify()
+            key.verify(sig, msg, self.padding, self.hash)
         except cryptography.exceptions.InvalidSignature as error:
             logger.debug(error, exc_info=True)
             return False
@@ -142,7 +125,7 @@ class _JWARSA:
 
 class _JWARS(_JWARSA, JWASignature):
 
-    def __init__(self, name, hash_):
+    def __init__(self, name: str, hash_: Callable[[], HashAlgorithm]) -> None:
         super().__init__(name)
         self.padding = padding.PKCS1v15()
         self.hash = hash_()
@@ -150,7 +133,7 @@ class _JWARS(_JWARSA, JWASignature):
 
 class _JWAPS(_JWARSA, JWASignature):
 
-    def __init__(self, name, hash_):
+    def __init__(self, name: str, hash_: Callable[[], HashAlgorithm]) -> None:
         super().__init__(name)
         self.padding = padding.PSS(
             mgf=padding.MGF1(hash_()),
@@ -161,11 +144,11 @@ class _JWAPS(_JWARSA, JWASignature):
 class _JWAEC(JWASignature):
     kty = jwk.JWKEC
 
-    def __init__(self, name, hash_):
+    def __init__(self, name: str, hash_: Callable[[], HashAlgorithm]):
         super().__init__(name)
         self.hash = hash_()
 
-    def sign(self, key, msg):
+    def sign(self, key: ec.EllipticCurvePrivateKey, msg: bytes) -> bytes:
         """Sign the ``msg`` using ``key``."""
         sig = self._sign(key, msg)
         dr, ds = decode_dss_signature(sig)
@@ -173,27 +156,17 @@ class _JWAEC(JWASignature):
         return (dr.to_bytes(length=length, byteorder='big') +
                 ds.to_bytes(length=length, byteorder='big'))
 
-    def _sign(self, key, msg):
-        # If cryptography library supports new style api (v1.4 and later)
-        new_api = hasattr(key, 'sign')
+    def _sign(self, key: ec.EllipticCurvePrivateKey, msg: bytes) -> bytes:
         try:
-            if new_api:
-                return key.sign(msg, ec.ECDSA(self.hash))
-            signer = key.signer(ec.ECDSA(self.hash))
+            return key.sign(msg, ec.ECDSA(self.hash))
         except AttributeError as error:
             logger.debug(error, exc_info=True)
             raise errors.Error('Public key cannot be used for signing')
         except ValueError as error:  # digest too large
             logger.debug(error, exc_info=True)
             raise errors.Error(str(error))
-        signer.update(msg)
-        try:
-            return signer.finalize()
-        except ValueError as error:
-            logger.debug(error, exc_info=True)
-            raise errors.Error(str(error))
 
-    def verify(self, key, msg, sig):
+    def verify(self, key: ec.EllipticCurvePublicKey, msg: bytes, sig: bytes) -> bool:
         """Verify the ``msg` and ``sig`` using ``key``."""
         rlen = jwk.JWKEC.expected_length_for_curve(key.curve)
         if len(sig) != 2 * rlen:
@@ -205,17 +178,9 @@ class _JWAEC(JWASignature):
         )
         return self._verify(key, msg, asn1sig)
 
-    def _verify(self, key, msg, asn1sig):
-        # If cryptography library supports new style api (v1.4 and later)
-        new_api = hasattr(key, 'verify')
-        if not new_api:
-            verifier = key.verifier(asn1sig, ec.ECDSA(self.hash))
-            verifier.update(msg)
+    def _verify(self, key: ec.EllipticCurvePublicKey, msg: bytes, asn1sig: bytes) -> bool:
         try:
-            if new_api:
-                key.verify(asn1sig, msg, ec.ECDSA(self.hash))
-            else:
-                verifier.verify()
+            key.verify(asn1sig, msg, ec.ECDSA(self.hash))
         except cryptography.exceptions.InvalidSignature as error:
             logger.debug(error, exc_info=True)
             return False
